import numpy as np
import sympy
import re
import sys
import copy

eq_attr_regex = r"\$\{([a-zA-Z_\/]+[a-zA-Z0-9_\/\(\)]*)\}"
equation_regex = "^(?P<eq>[\S\s]+)(?:\[)(\"(?P<name>[\S\s]*)\"\,){0,1}(?P<unit>[\S\s]*)(?:\])$"


def print_error(msg):
    sys.stderr.write(str(msg) + "\n")


def _eval_equation_list(data, equationString, do_lambdify = True):
    parsed_eq = _parse_main_equation(equationString)
    ds_list = parsed_eq["vars"]
    expression  = sympy.S(parsed_eq["equation"])

    if parsed_eq["success"] is False:
        print_error("Error: Failed to parse main equation to calculate data list")
        return np.array([])

    lengths = list(set(map(lambda x: len(data[x]), parsed_eq["vars"])))
    if len(lengths) > 1:
        print_error("Error: Lists involved in the equation are not equal in length")
        return np.array([])
    num_of_elements = lengths[0]


    #initialize the output array
    output_array    = np.empty(num_of_elements)
    output_array[:] = np.NaN

    if do_lambdify:
        try:
            lam_f = sympy.lambdify(tuple(ds_list), expression, modules='numpy')
        except:
            print_error("Error: Exception thrown while trying to lambdify the equation.")

        try:
            replacement_list = (np.array(data[ds_list[j]]) for j in range(len(ds_list)))
            output_array = np.double(lam_f(*replacement_list))
        except Exception as e:
            print_error("Error while evaluating equation expression. Exception says: " + str(e))

        is_nans_list = list(set(np.isnan(output_array)))
        if is_nans_list[0] == True and len(is_nans_list) == 1:
            print("Error: All equation output elements are NaN, which shows that the evaluation"
                  " of the equation failed. The output values will be replaced by zeros.")
            output_array = np.zeros(num_of_elements)
    else:

        for i in range(num_of_elements):
            rep_vals = {}
            for ds in ds_list:
                if re.match("___attr___(?P<attribute>[\w]+)", ds) is None:
                    rep_vals[ds] = data[ds][i]
                else:
                    rep_vals[ds] = data[ds][0]
            try:
                val = np.double(expression.subs(rep_vals))
            except Exception as e:
                print_error("Error while evaluating equation expression. Exception says: " + str(e))
                val = np.nan
            output_array[i] = val

    return output_array


def _parse_main_equation_multiple(eq_str):
    eq_str_split = eq_str.split("::")
    ret = []
    for eq in eq_str_split:
        ret.append(_parse_main_equation(eq))
    return ret


def _parse_main_equation(eq_str):
    """
    Parse the MainEquation and verify its validity
    Parameter: eqStr: The equation string
    Return: A dict with results. The element "success" of that dict
        is a boolean that indicates whether the parsing is success
    """
    if "__" in eq_str:
        print("Main equation cannot contain any variables with more than 1 underscore consequtively (_).")
        return {"success": False}

    def rep_ds_with_index_forward(s):
        # replace dataset's [[x]] with _____x_____ to avoid sympy replacement problems
        sv = str(s.group(0))
        sv = sv.replace("[[","")
        sv = sv.replace("]]","")
        return "_____" + sv + "_____"

    def rep_attr_forward(s):
        # replace attribute with ___attr___attribname to avoid sympy replacement problems
        sv = str(s.group(0))
        sv = sv.replace("${", "")
        sv = sv.replace("}", "")
        sv = sv.replace("/", "__")
        return "___attr___" + sv

    try:
        mt = re.match(equation_regex, eq_str)
        unit = mt.group("unit")
        name = mt.group("name")
        eq = mt.group("eq")
        eq = re.sub("\s+", "", eq)
        # replace for datasets with indices to indicate column number, so x[[5]]
        eq_tmp_reps = re.sub(r"\[\[\d*\]\]", rep_ds_with_index_forward, eq)
        # replace for attributes in the form ${dataset,attr}
        #eq_tmp_reps = re.sub(eq_attr_regex, rep_attr_forward, eq_tmp_reps)
        all_symbols = [str(x) for x in sympy.S(eq_tmp_reps).atoms(sympy.Symbol)]
        datasets = list(map(lambda x: re.sub(r"_____(\d)+_____", "", x), all_symbols))
        return {"success": True, "vars": all_symbols, "datasets": datasets, "units": unit,
                "equation": eq_tmp_reps, "original_equation": eq_str, "name": name}

    except Exception as e:
        print_error(e)
        return {"success": False}


# def GetHDF5FileData(fileObj,datasetsList):
#     """
#     Retrieves the data from a file.
#     The data can be a dataset in the form "dsName_____index_____" or
#             or an attribute of the form "___attr___attributeName___
#
#     Param: fileObj: HDF5 file object that's open
#     Param: datasetsList: List of string parameters to get
#     """
#     returnedData = {}
#     for ds in datasetsList:
#         only_ds = re.sub(r"_____\d+_____", "", ds)  #this doesn't contain indices
#         match_attr = re.match("___attr___(?P<dataset>[\w]+)__(?P<attribute>[\w]+)",only_ds)
#         match_attr_root = re.match("___attr___(?P<attribute>[\w]+)",only_ds)
#         if match_attr is not None:
#             attr_ds = match_attr.group("dataset")
#             attr_attr = match_attr.group("attribute")
#             returnedData[only_ds] = fileObj[attr_ds].attrs[attr_attr]
#         elif match_attr_root is not None:
#             attr_attr = match_attr_root.group("attribute")
#             returnedData[only_ds] = fileObj.attrs[attr_attr]
#         else:
#             if len(fileObj[only_ds].value.shape) == 1:
#                 ds_match = re.match(r"\S+_____\d+_____", ds)
#                 if ds_match is None:
#                     returnedData[ds] = fileObj[only_ds].value
#                 else:
#                     raise IndexError("Tried to choose an index for a single column dataset.")
#             elif len(fileObj[only_ds].value.shape) == 2: #if contains multiple columns, extract the column number
#                 ds_match = re.match(r"(?P<dsName>\S+)_____(?P<idx>\d*)_____", ds)
#                 if ds_match is None:
#                     raise re.error("Error: While the dimensionality of the data is 2, unable to match the requested variable (" + ds + ") with an index. Datasets provided: " + str(datasetsList))
#                 else:
#                     returnedData[ds] = fileObj[only_ds].value[:,int(ds_match.group("idx"))]
#
#             else:
#                 raise IndexError("Error: The dimensionality of the data in the dataset " + ds + " is not 1 or 2. This is not supported.")
#     return returnedData


def decode_string(str_in):
    if isinstance(str_in, bytes):
        ret_str = str_in.decode("ascii")
    else:
        ret_str = str_in
    return ret_str


def to_string(v):
    try:
        ret_val = v.astype(str)
    except AttributeError:
        ret_val = decode_string(v)
    return ret_val

def parse_equation_multiple(equations):
    equations_split = equations.split("::")
    ret = []
    for eq in equations_split:
        ret.append(parse_equation(eq))
    return ret

def parse_equation(equation):
    """
    Parse the MainEquation and verify its validity
    Parameter: equation: The equation string
    Return: A dict with results. The element "success" of that dict
        is a boolean that indicates whether the parsing is success.
        The following are expected returns in the dict:
        If "success" is False:
            "msg": Error message
        If "success" is True:
           "units": units parsed
           "name": name of the value parsed
           "datasets": datasets involved
           "datasets_indices": datasets involved, including indices in the form dataset_____idx_____
               where idx is the column number
           "attrs": attributes involved in the equation, if any
           "eq": The equation, stripped from units and name
    """
    attribs_list = []
    counter = [0]
    def rep_ds_with_index_forward(s):
        # replace dataset's [[x]] with _____x_____ to avoid sympy replacement problems
        sv = str(s.group(0))
        sv = sv.replace("[[","")
        sv = sv.replace("]]","")
        return "_____" + sv + "_____"

    def _rep_attr(s):
        # replace attribute with ___attr___attribname to avoid sympy replacement problems
        attrname = str(s.group(0))
        attrname = attrname.replace("${", "")
        attrname = attrname.replace("}", "")
        counter[0] += 1
        ret = "_____attr_____" + str(counter[0])
        attribs_list.append([attrname,ret])
        for i in range(len(attribs_list)-1):
            # replace recursive values that are named with temporary names
            attribs_list[-1][0] = attribs_list[-1][0].replace(attribs_list[i][1],attribs_list[i][0])
        return ret

    try:
        mt = re.match(equation_regex, equation)
        if mt is None:
            raise SyntaxError("Parsing equation failed. It doesn't follow the standard syntax.")
        unit = mt.group("unit")
        name = mt.group("name")
        eq = mt.group("eq")
        eq = re.sub("\s+", "", eq)  # remove spaces

        # recursively, substitute all attributes, until substitutions don't make any change
        prev_eq = eq
        curr_eq = eq
        while True:
            curr_eq = re.sub(eq_attr_regex, _rep_attr, prev_eq)
            # if no more substitutions change the equation, stop!
            if curr_eq == prev_eq:
                break
            else:
                prev_eq = curr_eq

        # replace for datasets with indices to indicate column number, so x[[5]]
        curr_eq = re.sub(r"\[\[\d*\]\]", rep_ds_with_index_forward, curr_eq)

        variables = [str(x) for x in sympy.S(curr_eq).atoms(sympy.Symbol)]
        variables = [x for x in variables if (re.match(r"_____attr_____\d+", x) is None)]  # remove attributes
        attribs_list = [x[0] for x in attribs_list]
        datasets = [re.sub(r"_____\d+_____", "", x) for x in variables]  # remove datasets indices
        return {"success": True, "units": unit, "name": name, "datasets": datasets, "datasets_indices": variables,
                "attrs": attribs_list, "equation": eq}

    except Exception as e:
        print_error("Failed to parse equation. Error: " + str(e))
        return {"success": False, "msg": str(e)}


def eval_equation_multiple(equations, do_lambdify = True, file=None,
                           dataset_getter=None, attr_getter=None):
    equations_split = equations.split("::")
    ret = []
    for eq in equations_split:
        ret.append(eval_equation(eq, do_lambdify, file,
                                          dataset_getter, attr_getter))
    return ret


def eval_equation(equation, do_lambdify = True, file = None,
                  dataset_getter = None, attr_getter = None):
    """
    Uses an HDF5 file object or dataset and attribute getters to evaluate an equation
    Format of the equation: Datasets as variables, attributes as ${attr}
        for global attributes and ${Dataset/attr} for attributes in datasets
    You can either specify a file, OR dataset_getter and attr_getter functions;
        these functions take 1 parameter only, which is the attribute name in
        the format dataset/attribute or a global attribute name. For datasets,
        if there are multiple columns, it should be supplied as dataset[i]
    Param: equation: Equation to be evaluated
    Param: file: HDF5 file object that's open
    """
    def _default_get_attr_from_file(attr):
        parts = attr.split("/")
        if len(parts) == 1:
            return to_string(file.attrs[parts[0]])
        elif len(parts) == 2:
            return to_string(file[parts[0]].attrs[parts[1]])
        else:
            raise SyntaxError("Only 1 level datasets are supported "
                              "for exctracting attributes. E.g.: "
                              "${dataset/attr} or ${attr}.")

    def _default_get_dataset_from_file(dataset):
        only_ds = re.sub(r"_____\d+_____", "", ds)  #this doesn't contain indices
        if len(file[only_ds].value.shape) == 1:
            ds_match = re.match(r"\S+_____\d+_____", ds)
            if ds_match is None:
                return copy.copy(file[only_ds].value)
            else:
                raise IndexError("Tried to choose an index for a single column dataset.")
        elif len(file[only_ds].value.shape) == 2: #if contains multiple columns, extract the column number
            ds_match = re.match(r"(?P<dsName>\S+)_____(?P<idx>\d*)_____", ds)
            if ds_match is None:
                raise re.error("Error: While the dimensionality of the data is 2, unable to match the requested variable (" + ds + ") with an index. Datasets provided: " + str(parsed_eq["vars"]))
            else:
                return copy.copy(file[only_ds].value[:,int(ds_match.group("idx"))])
        else:
            raise IndexError("Error: The dimensionality of the data in the dataset " + ds + " is not 1 or 2. This is not supported.")

    if file is None and dataset_getter is None and attr_getter is None:
        print_error("You should either set a file object or dataset and attribute getters")
        return []

    if file is not None:
        attr_getter_internal = _default_get_attr_from_file
        dataset_getter_internal = _default_get_dataset_from_file
    else:
        if dataset_getter is None and attr_getter is None:
            raise IOError("You should either specify a file, or "
                          "dataset and attribute getters")
        else:
            dataset_getter_internal = dataset_getter
            attr_getter_internal = attr_getter

    def _rep_attr(s):
        # replace attribute's format ${attrib} with ___attr___attribname to avoid sympy replacement problems
        if attr_getter_internal is None:
            raise IOError("Requested attributes without setting attr_getter"
                          " or setting a file object")
        sv = str(s.group(0))
        sv = sv.replace("${", "")
        sv = sv.replace("}", "")
        try:
            ret = attr_getter_internal(sv)
            return ret
        except Exception as e:
            print_error("Attribute substitution failed. Exception says: " + str(e))
            raise

    # recursively, substitute all attributes
    prev_eq = equation
    curr_eq = equation
    while True:
        curr_eq = re.sub(eq_attr_regex, _rep_attr, prev_eq)
        # if no more substitutions change the equation, stop!
        if curr_eq == prev_eq:
            break
        else:
            prev_eq = curr_eq
    returned_data = {}
    parsed_eq = parse_equation(curr_eq)
    if not parsed_eq["success"]:
        raise SyntaxError("Parsing equation failed. Error says: " + str(parsed_eq["msg"]))

    if len(parsed_eq["datasets_indices"]) != 0:
        if dataset_getter_internal is None:
            raise IOError("Requested datasets without setting dataset_getter"
                          " or setting a file object")

    for ds in parsed_eq["datasets_indices"]:
        try:
            returned_data[ds] = dataset_getter_internal(ds)
        except Exception as e:
            print_error("Dataset pulling failed. Exception says: " + str(e))
            raise

    return _eval_equation_list(returned_data, curr_eq, do_lambdify)



from GNOMEAcquirerLibs.Dataset import *


# This class is made to concatenate data from DataBatch
class AcquisitionConfig():
    def __init__(self):
        self.requestedDevice = None
        self.requestedChannels = {}
        self.requestedChannelsTypes = {}
        # self.conversionFactors = {}
        # self.chOffsets = {}
        self.isFakeAcquisition = False

        self.isFileAcquisition = False
        self.fileAcqPath       = None

        self.errorFlag = False
        self.errorMessage = ""

        self.dataPath = ""
        self.stationName = ""
        self.attributes = {}
        self.paranoidMode = False

    #set device to be used
    def registerDevice(self,acquisitionWindow):
        self.requestedDevice = acquisitionWindow.portLine.currentText()

    def registerChannels(self,acquisitionWindow):
        self.requestedChannels = {}
        self.requestedChannelsTypes = {}

        for tabNum in range(0,acquisitionWindow.datasetsTabs.count()):
            #get channel numbers
            tmpChNum = int(acquisitionWindow.datasetsTabs.widget(tabNum).channelLine.currentText())
            tmpChType = acquisitionWindow.datasetsTabs.widget(tabNum).channelDatatypeLine.currentText()
            if tmpChNum in self.requestedChannels.values():
                self.errorFlag = True
                self.errorMessage = "Error: You cannot request the same channel for two datasets"
                return
            self.requestedChannels[acquisitionWindow.datasetsTabs.tabText(tabNum)] = tmpChNum
            self.requestedChannelsTypes[acquisitionWindow.datasetsTabs.tabText(tabNum)] = tmpChType



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import sys
import os

if hasattr(sys, 'frozen'):
    ScriptPath = os.path.join(os.path.dirname(sys.executable), "GNOMEAcquirerLibs")
else:
    ScriptPath = os.path.dirname(os.path.realpath(__file__))


class ViewLog(QWidget):
    def __init__(self, parent = None):
        super(ViewLog, self).__init__(parent)

        #set window icon

        self.windowIcon = QIcon(os.path.join(ScriptPath, "LogFile.png"))
        self.setWindowIcon(self.windowIcon)
        self.setWindowTitle("Log file")

        self.lastLogFilePath = ""
        self.log_oldContent = ""
        self.log_newContent = ""

        self.mainLayout = QGridLayout()
        self.setLayout(self.mainLayout)

        self.mainGroupBox = QGroupBox("Log view")
        self.logTextField = QTextEdit()
        self.logTextField.setReadOnly(1)
        self.logFilePathLineEdit = QLineEdit()
        self.logFilePathLineEdit.setReadOnly(1)
        self.logFilePathLineEdit.setToolTip("Log file path")
        self.reloadButton = QPushButton("Reload log file")
        self.reloadButton.clicked.connect(self.reloadLogFile)

        self.groupBoxLayout = QGridLayout()
        self.groupBoxLayout.addWidget(self.logFilePathLineEdit,0,0,1,1)
        self.groupBoxLayout.addWidget(self.logTextField, 1, 0, 1, 1)
        self.groupBoxLayout.addWidget(self.reloadButton, 2, 0, 1, 1)

        self.mainGroupBox.setLayout(self.groupBoxLayout)
        self.mainLayout.addWidget(self.mainGroupBox)

        self.setWindowState(Qt.WindowMaximized)


    def setLogFile(self,logFilePath):
        if logFilePath == "":
            QMessageBox.information(self,"Error","There is still no log file. You have to start the acquisition first.")
            return -1
        try:
            with open(logFilePath, 'r') as f:
                logFileData = f.readlines()
            self.logFilePathLineEdit.setText(logFilePath)
            self.log_oldContent = self.log_newContent
            self.log_newContent = "".join(logFileData)
            if self.log_newContent != self.log_oldContent:
                self.logTextField.setText(self.log_newContent)
                self.logTextField.verticalScrollBar().setValue(self.logTextField.verticalScrollBar().maximum())
            self.lastLogFilePath = logFilePath

            return 0
        except Exception as e:
            QMessageBox.information(self,"Error","There was a problem while trying to open the log file \"" + logFilePath + "\". Are you sure you haven't deleted the log file? The error says: " + str(e))
            return -2

    def reloadLogFile(self):
        self.setLogFile(self.lastLogFilePath)


#!/usr/bin/python3

# import cProfile
import multiprocessing
from GNOMEAcquirerLibs.AcquisitionWindow import *

# enable this to test frozen version of the program
# setattr(sys, 'frozen', True)


def main_function():
    multiprocessing.freeze_support()

    PrintSoftwareInfo()

    app = QApplication(sys.argv)

    # make message boxes text selectable
    app.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }");

    screen = AcqusitionWindow()
    screen.showMaximized()
    screen.doAfterShown()

    app.aboutToQuit.connect(screen.cleanUp)

    sys.exit(app.exec_())

# if __name__ == '__main__':
if __name__.endswith('__main__'):  # fix because cx_freeze changes __name__
    main_function()
    # cProfile.run('main_function()',sort='tottime')



import datetime
import copy
from GNOMEAcquirerLibs.AttributeData import *

# constants
# The following define error messages definitions, as communicated between the main thread and the acquisition thread
ERROR_SEVERITY_SEVERE  = "Severe"
ERROR_SEVERITY_WARNING = "Warning"

ERROR_IDENTIFIER_SEVERITY = "Severity"
ERROR_IDENTIFIER_MESSAGE  = "Message"


def GetChannelToArrayMapping(channelOpBinary):
    """
    Channel to array mapping is necessary because if a channel is disabled, consequently a column disappears instead
    of making it empty.
    Therefore, it's important to detect what channels are enabled, and then map them in order with the available column.
    ChannelToArray means channel number in the device vs the column number in the array
    :param channelOpBinary: A list with length that is equal to the maximum number of channels (so, 4).
        It contains as boolean flags whether a every channel is enabled
    :return: Returns a dict that has for every enabled channel the corresponding column in the array. If the channel
    doesn't exist, its key doesn't exist in the dict.
    """
    #find channel to array element correspondence. In the acquisition box, if a channel is disabled, it doesn't produce output
    channelToArrayMapping = {}
    chToArrCounter = 0
    #check available channels
    for i in range(len(channelOpBinary)):
        if channelOpBinary[i]:
            channelToArrayMapping[i] = chToArrCounter
            chToArrCounter += 1
    return channelToArrayMapping


def FindDataRange(StrQueue, startStr, endStr):
    startIndex = next((i for i in range(len(StrQueue)) if StrQueue[i].replace(" ","").replace("\t","") == startStr), None)
    if startIndex is not None:
        endIndex = next(
            (i for i in range(startIndex + 1, len(StrQueue)) if StrQueue[i] == endStr), None)
        if endIndex is None:
            raise LookupError("Cannot find " + endStr)
    else:
        raise LookupError("Cannot find " + startStr)
    return startIndex, endIndex+1


# DataBatch is a class whose object contains the data received from the device in a single shot.
# DataBatch consists mainly of a string queue that is received from the device (we call batch here).
# This string queue is analyzed based on a certain order of the information.
class DataBatch:
    SamplingRatePossibilities = [20,50,100,200,500,1000]
    d_dataPrefix       = "@Data"
    d_magneticPrefix   = "@Magnetic"
    d_startString      = "@Header"
    d_endString        = "@End"
    d_timePrefix       = "Time:"
    d_datePrefix       = "Date:"
    d_intTempPrefix    = "Temperature internal [C]:"
    d_extTempPrefix    = "Temperature external [C]:"
    d_latitudePrefix   = "Latitude [deg]:"
    d_longitudePrefix  = "Longitude [deg]:"
    d_altitudePrefix   = "Altitude [m]:"
    d_weekNumberPrefix = "Week number:"
    d_timeNotSetStr = "Time not set!"
    f_dateFormat = "%Y.%m.%d"
    f_timeFormat = "%H.%M.%S"

    def __init__(self, acquisitionConfig, dataQueueOfStrings):
        self.deviceName       = acquisitionConfig.requestedDevice
        self.dataStringQueue  = dataQueueOfStrings
        self.requestedChannelsInDev = acquisitionConfig.requestedChannels
        self.requestedChannelsTypesInDev = acquisitionConfig.requestedChannelsTypes

        self.dataArray        = []
        #Our acquisition system has a built-in (auxiliary) sensor for magnetic fields. These are stored here.
        self.auxMagFieldSensorArray = []
        self.samplingRate     = None
        self.missingPoints    = None
        self.startTime        = None
        self.weekNumber       = None
        self.latitude         = None
        self.longitude        = None
        self.altitude         = None
        self.channelOp        = []
        self.channelOpBinary  = []
        self.receiverMode     = None
        self.intTemperature   = None
        self.extTemperature   = None
        self.magFieldSenorUnit = ""
        self.headerStringList = []
        self.errorFlag = False
        self.errorMsgs = []

        self.parseAllData()


    def containsSevereError(self):
        for err in self.errorMsgs:
            if err[ERROR_IDENTIFIER_SEVERITY] == ERROR_SEVERITY_SEVERE:
                return err
        return None

    def parseAllData(self):
        self.checkParameterInstanceSanity(DataBatch.d_timePrefix)
        self.checkParameterInstanceSanity(DataBatch.d_datePrefix)
        self.checkParameterInstanceSanity(DataBatch.d_intTempPrefix)
        self.checkParameterInstanceSanity(DataBatch.d_extTempPrefix)
        self.checkParameterInstanceSanity(DataBatch.d_latitudePrefix)
        self.checkParameterInstanceSanity(DataBatch.d_longitudePrefix)
        self.checkParameterInstanceSanity(DataBatch.d_altitudePrefix)
        self.checkParameterInstanceSanity(DataBatch.d_weekNumberPrefix)
        if self.errorFlag:
            return self.errorFlag

        if not self.containsSevereError():
            self.parseStartTime()
        if not self.containsSevereError():
            self.parseDataArray()
        if not self.containsSevereError():
            self.parseAuxSensorMagneticFields()
        if not self.containsSevereError():
            self.checkChannelSelectionSanity()
        if not self.containsSevereError():
            self.castDataToCorrectTypes()
        if not self.containsSevereError():
            self.parseCoordinates()
        if not self.containsSevereError():
            self.parseTemperature()
        if not self.containsSevereError():
            self.parseWeekNumber()
        if not self.containsSevereError():
            self.calculateSamplingRate()

        if self.errorFlag:
            return self.errorFlag



    #search string data queue and find a parameter by its prefix.
    #returns a list of lines containing these instances
    def findPrefixInDataQueue(self,prefix):
        return [line for line in self.dataStringQueue if line[:len(prefix)] == prefix]

    #checks whether every parameter exists in every batch only once!
    #a parameter is defined by its prefix in a line
    def checkParameterInstanceSanity(self,prefix):
        dataInstancesCount = sum(line[:len(prefix)] == prefix for line in self.dataStringQueue)
        if dataInstancesCount > 1:
            self.setErrorFlag("Multiple instances of \"" + prefix + "\" string was found in a single data batch. " \
                                                                          "The current model of data " \
                                                                          "includes that only once",ERROR_SEVERITY_SEVERE)
        elif dataInstancesCount == 0:
            self.setErrorFlag("Could not find \"" + prefix + "\" in the current data batch. " \
                                                                   "The current model of the data contains " \
                                                                   "that once per batch.",ERROR_SEVERITY_SEVERE)

    def parseStartTime(self):
        try:
            self.timeStrList = self.findPrefixInDataQueue(self.d_timePrefix)
            self.dateStrList = self.findPrefixInDataQueue(self.d_datePrefix)

            #remove tabs and spaces
            self.timeStr = self.timeStrList[0]
            self.dateStr = self.dateStrList[0]

            # Check if the last part of time equals the timeNotSetStr string, which indicates that GPS is not connected.
            # This part is written after the time string
            if len(self.timeStr) >= len(DataBatch.d_timeNotSetStr):
                if self.timeStr[-len(DataBatch.d_timeNotSetStr):] == DataBatch.d_timeNotSetStr:
                    self.timeStr = self.timeStr[:-len(DataBatch.d_timeNotSetStr)].replace(" ","").replace("\t","")
                    self.setErrorFlag("Time is claimed to be not set. Antenna has to be connected for the data acquired to be meaningful")
                    #return

            self.timeStr = self.timeStr.replace(" ","").replace("\t","")
            self.dateStr = self.dateStr.replace(" ","").replace("\t","")

            #remove strings prefix
            timeStrToParse = self.timeStr.replace(DataBatch.d_timePrefix,"")
            dateStrToParse = self.dateStr.replace(DataBatch.d_datePrefix,"")

            sep = "___"

            try:
                self.startTime = datetime.datetime.strptime(timeStrToParse + sep + dateStrToParse,DataBatch.f_timeFormat + sep + DataBatch.f_dateFormat)
            except Exception as e:
                self.setErrorFlag("Times and Dates are supposed to contain three, two digit numbers separated by a dot, \".\", " \
                                 "This does not seem to be the case! " \
                                 "Check the syntax of the received data from the device. The string seen as time is: " +
                                 "Time: " + str(timeStrToParse) + " | " + "Date: " +str(dateStrToParse) + ". Exception message: " + str(e)
                                  ,ERROR_SEVERITY_SEVERE)
                return

        except Exception as e:
            self.setErrorFlag("An exception was thrown while parsing time. Make sure that the device is sending correct time format." + " Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
            return

    def parseCoordinates(self):
        try:
            longitudeStr = self.findPrefixInDataQueue(self.d_longitudePrefix)[0]
            latitudeStr  = self.findPrefixInDataQueue(self.d_latitudePrefix)[0]
            altitudeStr  = self.findPrefixInDataQueue(self.d_altitudePrefix)[0]

            self.longitude = float(longitudeStr[len(self.d_longitudePrefix):].replace(" ","").replace("\t",""))
            self.latitude  = float(latitudeStr[len(self.d_longitudePrefix):].replace(" ","").replace("\t",""))
            self.altitude  = float(altitudeStr[len(self.d_longitudePrefix):].replace(" ","").replace("\t",""))

        except Exception as e:
            self.setErrorFlag("Could not find/convert coordinates (longitude,latitude,altitude)." + " Exception message: " + str(e),ERROR_SEVERITY_SEVERE)

    def parseTemperature(self):
        try:
            intTempStr = self.findPrefixInDataQueue(self.d_intTempPrefix)[0]
            extTempStr = self.findPrefixInDataQueue(self.d_extTempPrefix)[0]

            self.intTemperature = str(intTempStr[len(self.d_intTempPrefix):].replace(" ","").replace("\t",""))
            self.extTemperature = str(extTempStr[len(self.d_extTempPrefix):].replace(" ","").replace("\t",""))
            try:
                self.intTemperature = float(self.intTemperature)
            except:
                self.intTemperature = "nan"

            try:
                self.extTemperature = float(self.extTemperature)
            except:
                self.extTemperature = "nan"


        except Exception as e:
            self.setErrorFlag("Could not find/convert temperature information." + " Exception message: " + str(e))

    def parseWeekNumber(self):
        try:
            weekNumberStr = self.findPrefixInDataQueue(self.d_weekNumberPrefix)[0]
            self.weekNumber = int(weekNumberStr[len(self.d_weekNumberPrefix):].replace(" ","").replace("\t",""))

        except Exception as e:
            self.setErrorFlag("Could not find/convert week number." + " Exception message: " + str(e), ERROR_SEVERITY_SEVERE)


    def parseDataArray(self):
        try:
            range = FindDataRange(self.dataStringQueue, DataBatch.d_dataPrefix, DataBatch.d_magneticPrefix)
        except Exception as e:
            self.setErrorFlag("Error while searching for auxiliary sensor magnetic field data."
                              + " Exception message: " + str(e))


        self.parseChannelStatusString(self.dataStringQueue[range[0]+1])

        try:
            self.dataArray = list(map(str.split,self.dataStringQueue[range[0]+2:range[1]-1]))
        except Exception as e:
            self.setErrorFlag("Error while pushing acquired data points into an array." + " Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
            return

        # self.getChannelStateBits = False
        # self.startCollecting = False
        # for line in self.dataStringQueue:
        #     if line.replace(" ","").replace("\t","") == DataBatch.d_dataPrefix:
        #         self.getChannelStateBits = True
        #         continue
        #
        #     if self.getChannelStateBits:
        #         self.parseChannelStatusString(line)
        #         self.getChannelStateBits = False
        #         self.startCollecting = True
        #         continue
        #
        #     if self.startCollecting:
        #         if line.replace(" ","").replace("\t","") == DataBatch.d_magneticPrefix:
        #             break
        #         else:
        #             try:
        #                 self.splitLine = line.split()
        #                 self.dataArray.append(self.splitLine)
        #             except Exception as e:
        #                 self.setErrorFlag("Error while pushing acquired data points into an array." + " Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
        #                 return

    def parseAuxSensorMagneticFields(self):

        try:
            range = FindDataRange(self.dataStringQueue, DataBatch.d_magneticPrefix, DataBatch.d_endString)
        except Exception as e:
            self.setErrorFlag("Error while searching for auxiliary sensor magnetic field data."
                              + " Exception message: " + str(e))

        try:
            self.parseAuxMagneticFieldsStatus(self.dataStringQueue[range[0]+1])
            self.auxMagFieldSensorArray = list(map(lambda p: np.array(list(map(np.float32,p.split()))),self.dataStringQueue[range[0]+2:range[1]-1]))

        except Exception as e:
            self.setErrorFlag("Error while pushing auxiliary sensor magnetic field data points into an array."
                              " It is possible that the sensor is not properly connected." + " Exception message: " + str(e))
            return

        # self.getMagStateBits = False
        # self.startCollecting = False
        # for line in self.dataStringQueue:
        #     if line.replace(" ","").replace("\t","") == DataBatch.d_magneticPrefix:
        #         self.getMagStateBits = True
        #         continue
        #
        #     if self.getMagStateBits:
        #         self.parseAuxMagneticFieldsStatus(line)
        #         self.getMagStateBits = False
        #         self.startCollecting = True
        #         continue
        #
        #     if self.startCollecting:
        #         if line.replace(" ","").replace("\t","") == DataBatch.d_endString:
        #             break
        #         else:
        #             try:
        #                 self.splitLine = list(map(np.float32,line.split()))
        #                 self.auxMagFieldSensorArray.append(np.array(self.splitLine))
        #             except Exception as e:
        #                 self.setErrorFlag("Error while pushing auxiliary sensor magnetic field data points into an array."
        #                                   " It is possible that the sensor is not properly connected." + " Exception message: " + str(e))
        #                 return

    def castDataToCorrectTypes(self):
        self.dataArrayTranspose = list(map(list, list(zip(*self.dataArray))))

        self.channelToArrayMapping = GetChannelToArrayMapping(self.channelOpBinary)

        self.dataToSave = {}
        self.channelRangesPerName = {}
        for dataset in self.requestedChannelsInDev:
            try:
                dataArrayIndex = self.channelToArrayMapping[self.requestedChannelsInDev[dataset]-1]
                self.dataToSave[dataset] = self.dataArrayTranspose[dataArrayIndex]

                #save channel ranges from channel name (instead of number)
                self.channelRangesPerName[dataset] = self.channelRanges[self.requestedChannelsInDev[dataset]-1]
            except Exception as e:
                print("Error: Exception with: " + dataset + ": ",e)
                self.setErrorFlag("Error while resorting data from recorded arrays. Make sure you enabled the channels "
                                  "you requested from the acquisition box. Check the data log to find the problem in the data." + " Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
                return
            try:
                if self.requestedChannelsTypesInDev[dataset] == GetTypeNameString(AttributeType.Double):
                    typeToCast = np.float64
                elif self.requestedChannelsTypesInDev[dataset] == GetTypeNameString(AttributeType.Float):
                    typeToCast = np.float32
                elif self.requestedChannelsTypesInDev[dataset] == GetTypeNameString(AttributeType.Integer32):
                    typeToCast = np.int32
                elif self.requestedChannelsTypesInDev[dataset] == GetTypeNameString(AttributeType.Integer64):
                    typeToCast = np.int64
                elif self.requestedChannelsTypesInDev[dataset] == GetTypeNameString(AttributeType.String):
                    typeToCast = np.str
                else:
                    print("Dataset received: " + str(self.dataToSave[dataset]))
                    dataSampleForException = self.dataToSave[dataset][0:4] if len(self.dataToSave[dataset]) > 4 else self.dataToSave[dataset]
                    self.setErrorFlag("No supported data type was found to cast dataset " + dataset + " to " + self.requestedChannelsTypesInDev[dataset] + ". Check the terminal log of the program to see what the full data array looks like. Part of the data is " + str(dataSampleForException) + ". Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
                    print(self.dataToSave)

                # self.dataToSave[dataset] = (np.asarray(self.dataToSave[dataset],dtype=typeToCast) + typeToCast(self.chOffsets[dataset]))*typeToCast(self.conversionFactors[dataset])
                self.dataToSave[dataset] = (np.asarray(self.dataToSave[dataset],dtype=typeToCast))

            except Exception as e:
                print(self.requestedChannelsInDev)
                print(self.dataToSave)
                print("Dataset received: " + str(self.dataToSave[dataset]))
                print("Exception message: ",e)
                dataSampleForException = self.dataToSave[dataset][0:4] if len(self.dataToSave[dataset]) > 4 else self.dataToSave[dataset]
                self.setErrorFlag("Error while trying to cast data of dataset " + dataset + " to " + self.requestedChannelsTypesInDev[dataset] + ". Check the terminal log of the program to see what the full data array looks like. Part of the data is " + str(dataSampleForException) + ". Exception message: " + str(e),ERROR_SEVERITY_SEVERE)


    def parseChannelStatusString(self,chString):
        self.splitChannelStatus = chString.split(",")
        if not len(self.splitChannelStatus) == 4:
            self.setErrorFlag("The line coming directly after " + DataBatch.d_dataPrefix + " is expected"
                              " to contain 4 strings separated by commas. This was not found. Unknown error!",ERROR_SEVERITY_SEVERE)
            return

        else:
            for ch in self.splitChannelStatus:
                chStripped = ch.replace(" ","").replace("\t","")
                #the following removes the prefix strings of channels status, i.e., Ch1, for example.
                self.channelOp.append(chStripped[3:])
        #convert channel findings to binary values
        self.channelOpToBin()
        self.channelOpToRanges()
        # print(self.channelOpBinary)

    def parseAuxMagneticFieldsStatus(self,magString):
        self.magFieldSenorStatus = magString.split()
        if len(self.magFieldSenorStatus) < 4:
            self.setErrorFlag("The header of sensor magnetic fields that comes after " + DataBatch.d_magneticPrefix + " is expected"
                              " to contain 4 or more strings separated by a space. This was not found. "
                              "The string seen is: \"" + magString + "\".", ERROR_SEVERITY_SEVERE)
            return

        else:
            #the range comes 4th
            self.magFieldSenorRange = self.magFieldSenorStatus[3]
            if len(self.magFieldSenorStatus) > 4:
                self.magFieldSenorUnit  = self.magFieldSenorStatus[4].replace("[","").replace("]","")
        # print(self.channelOpBinary)


    #write channel states as booleans
    def channelOpToBin(self):
        #self.channelOp contains whether channels are on or off as strings, it's either "off" or a range
        self.channelOpBinary = list(map(lambda channelStrStatus: True if not channelStrStatus == "off" else False, self.channelOp))

    def channelOpToRanges(self):
                #initialize an array with garbage (meaningless) values. This is because python doesn't have a resize function
        #in its default array
        self.channelRanges = self.channelOp

    def checkChannelSelectionSanity(self):
        self.channelToArrayMapping = GetChannelToArrayMapping(self.channelOpBinary)
        channelCount = 0
        for dataset in self.requestedChannelsInDev:
            binVal = self.channelOpBinary[channelCount]
            if binVal:
                if not (self.requestedChannelsInDev[dataset]-1 in self.channelToArrayMapping):
                    self.setErrorFlag("Error in channel presence. Channel " + str(self.requestedChannelsInDev[dataset]) + " in device " + str(self.deviceName) + " is not enabled and/or does not exist in the acquired data.",ERROR_SEVERITY_SEVERE)
                    print("Error in channel selection. Channel " + str(self.requestedChannelsInDev[dataset]) + " in device " + str(self.deviceName) + " is not enabled and/or does not exist in the acquired data.")

            channelCount += 1

    def calculateSamplingRate(self):
        if len(self.dataArray) < 15:
            self.setErrorFlag("The sampling rate of the data cannot be calculated before parsing the data points. "
                              "Or: You cannot choose a sampling rate that is less than 20 Hz. This is due to the algorithm "
                              "used to calculate the sampling rate and determine whether there are missing points in the data.",ERROR_SEVERITY_SEVERE)
        else:
            try:
                #the following is an algorithm that detects the existence of missing points by comparing with the possible
                #values of the sampling rate
                self.dataLength = len(self.dataArray)
                self.ident = list(map(lambda x: abs(x - self.dataLength), DataBatch.SamplingRatePossibilities))
                self.missingPoints = min(self.ident)
                self.samplingRate = DataBatch.SamplingRatePossibilities[self.ident.index(self.missingPoints)]
                if self.missingPoints != 0:
                    self.setErrorFlag("WARNING: There are missing data points in one of the data batches. Number of missing points is: " + str(self.missingPoints),severity=ERROR_SEVERITY_SEVERE)
            except Exception as e:
                self.setErrorFlag("An exception was thrown while trying to calculate the sampling rate. Exception message: " + str(e),ERROR_SEVERITY_SEVERE)
                print("An exception was thrown while trying to calculate the sampling rate. Exception message: " + str(e))

    def setErrorFlag(self,message,severity=ERROR_SEVERITY_WARNING):
        self.errorFlag = True
        self.errorMsgs.append({ERROR_IDENTIFIER_MESSAGE:message, ERROR_IDENTIFIER_SEVERITY:severity})



import copy
import os
import datetime
import numpy as np
import h5py
import sys
from GNOMEAcquirerLibs.Dataset import *
from GNOMEAcquirerLibs.AttributeData import *
from GNOMEAcquirerLibs.DataBatch import *
from GNOMEAcquirerLibs.Meta import *

def CreateDir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)

BatchesToStoredInSingleFile = 60

CorruptDataSubfolder = "CorruptData"

logDateTimeFormat = "%Y-%m-%d_%H-%M-%S_UTC"
logDirName        = "Logs"
logFileNamePrefix = "AcquisitionLog_"
logFileNameSuffix = ".txt"

def ApplyAttributeReplacement(attrStr,data,acqConfig,datasetName):
    attrStrReplaced = copy.deepcopy(attrStr)
    StartTime    = data.getStartTimeString()
    EndTime      = data.getEndTimeString()
    Date         = data.getDateString()
    SR           = data.getSamplingRateString()
    Weeknum      = data.getWeekNumberString()
    Longitude    = data.getLongitudeString()
    Latitude     = data.getLatitudeString()
    Altitude     = data.getAltitudeString()
    IntTemp      = data.getIntTemperatureString()
    ExtTemp      = data.getExtTemperatureString()
    ChRange      = data.getChannelRangeString(datasetName)
    MissingPoints = data.getMissingPointsString()

    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_StartTime,StartTime)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_EndTime,EndTime)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Date,Date)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_SamplingRate,SR)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_SamplingRate_SanityChannel,"1")
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_WeekNumber,Weeknum)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Altitude,Altitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Latitude,Latitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_Longitude,Longitude)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_InternalTemperature,IntTemp)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_ExternalTemperature,ExtTemp)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_ChannelRange,ChRange)
    attrStrReplaced = attrStrReplaced.replace(AttributeKeyword_MissingPoints,MissingPoints)

    return attrStrReplaced

# This class is made to concatenate data from DataBatch. It contains the mechanism to concatenate data batches together
# to create a dataset. The most important function here is DataCollected.appendData()
class DataCollected():
    BatchSecondsLength = 1
    #whether there were problems when adding data
    AppendData_ReturnSuccessKey   = "Success"
    #error message when problems exist
    AppendData_ReturnErrorMsgKey  = "ErrorMsg"
    #whether it's the last batch of a single file
    AppendData_ReturnIsLastBatchKey = "LastBatch"

    def __init__(self, acqConfig):
        self.acqConfig = acqConfig
        self.initialize()

    def setAttributes(self, attrs):
        self.acqConfig.attributes = attrs

    def initialize(self):
        self.dataArray = {}
        self.auxMagFieldSensorArray = np.array([])
        self.intTemperatureArray = np.array([])
        self.extTemperatureArray = np.array([])
        self.numOfPoints = 0
        self.missingPoints = 0
        self.samplingRate = 0
        self.startTime = None
        self.numOfBatches = 0
        self.lastAddedBatchTime = None

    def appendData(self,dataBatch):
        if len(self.dataArray) == 0:
            self.dataArray    = copy.deepcopy(dataBatch.dataToSave)
            self.auxMagFieldSensorArray = np.array(dataBatch.auxMagFieldSensorArray)
            self.magFieldSenorRange = dataBatch.magFieldSenorRange
            self.magFieldSenorUnit  = dataBatch.magFieldSenorUnit
            self.channelRangesPerName    = dataBatch.channelRangesPerName

            self.startTime    = dataBatch.startTime
            self.samplingRate = dataBatch.samplingRate
            self.numOfBatches = 1
            self.numOfPoints  = len(dataBatch.dataArray)
            self.weekNumber   = dataBatch.weekNumber
            self.latitude     = dataBatch.latitude
            self.longitude    = dataBatch.longitude
            self.altitude     = dataBatch.altitude

            self.intTemperatureArray = np.array([])
            self.extTemperatureArray = np.array([])
            self.intTemperature = dataBatch.intTemperature
            self.intTemperatureArray = np.concatenate([self.intTemperatureArray, [np.float32(dataBatch.intTemperature)] ])
            self.extTemperature = dataBatch.extTemperature
            self.extTemperatureArray = np.concatenate([self.extTemperatureArray, [np.float32(dataBatch.extTemperature)] ])

            self.errorFlag = dataBatch.errorFlag
            self.errorMsgs = dataBatch.errorMsgs

            self.lastAddedBatchTime = dataBatch.startTime

            self.checkNumberOfBatches()

            return {DataCollected.AppendData_ReturnSuccessKey  : True,
                    DataCollected.AppendData_ReturnErrorMsgKey : "",
                    DataCollected.AppendData_ReturnIsLastBatchKey: (self.numOfBatches == 0)}

        else:
            self.missingPoints += dataBatch.missingPoints
            self.numOfBatches += 1
            appendedLength = 0
            self.intTemperatureArray = np.concatenate([self.intTemperatureArray, [np.float32(dataBatch.intTemperature)] ])
            self.extTemperatureArray = np.concatenate([self.extTemperatureArray, [np.float32(dataBatch.extTemperature)] ])

            for dataset in dataBatch.dataToSave:
                self.dataArray[dataset] = np.concatenate([self.dataArray[dataset], dataBatch.dataToSave[dataset]])
                appendedLength = len(dataBatch.dataToSave[dataset])

            if len(dataBatch.auxMagFieldSensorArray) > 0:
                if len(self.auxMagFieldSensorArray) > 0:
                    self.auxMagFieldSensorArray = np.concatenate([self.auxMagFieldSensorArray,np.array(dataBatch.auxMagFieldSensorArray)])
                else:
                    self.auxMagFieldSensorArray = np.array(dataBatch.auxMagFieldSensorArray)
            self.numOfPoints += appendedLength
            self.errorFlag = self.errorFlag or dataBatch.errorFlag
            self.errorMsgs = self.errorMsgs + dataBatch.errorMsgs
            #remove duplicates
            self.errorMsgs = [dict(t) for t in set([tuple(d.items()) for d in self.errorMsgs])]

            #check for that batches are consecutive
            if not (self.lastAddedBatchTime is None):
                timeDiff = dataBatch.startTime - self.lastAddedBatchTime
                if abs(timeDiff.total_seconds()-1.0) >= 1e-6:
                    return {DataCollected.AppendData_ReturnSuccessKey: False,
                            DataCollected.AppendData_ReturnErrorMsgKey: "Error while adding batches from the received data (time of batch: " + str(self.startTime) + ". The data received is not consecutive in time.\n"
                                "The difference in time between the last two batches is: " + str(timeDiff.total_seconds()) + " seconds",
                            DataCollected.AppendData_ReturnIsLastBatchKey: (self.numOfBatches == 0)}

                else:
                    self.lastAddedBatchTime = dataBatch.startTime
                    self.checkNumberOfBatches()
                    return {DataCollected.AppendData_ReturnSuccessKey: True,
                            DataCollected.AppendData_ReturnErrorMsgKey: "",
                            DataCollected.AppendData_ReturnIsLastBatchKey: (self.numOfBatches == 0)}

    def checkNumberOfBatches(self):
        if self.numOfBatches >= BatchesToStoredInSingleFile:
            self.dumpDataToFile()
            self.clearData()

    def length(self):
        return self.numOfPoints

    def size(self):
        return self.length()

    def getLongitudeString(self):
        return str(self.longitude)

    def getLatitudeString(self):
        return str(self.latitude)

    def getAltitudeString(self):
        return str(self.altitude)

    def getStartTimeString(self):
        tmpTime = self.startTime
        return str(str(tmpTime.hour).zfill(2) + ":" + str(tmpTime.minute).zfill(2) + ":" + str(tmpTime.second).zfill(2) + "." + str(tmpTime.microsecond).zfill(6))

    def getEndTimeString(self):
        tmpTime = self.startTime
        tmpTime = tmpTime + datetime.timedelta(seconds = self.numOfBatches*DataCollected.BatchSecondsLength)
        return str(str(tmpTime.hour).zfill(2) + ":" + str(tmpTime.minute).zfill(2) + ":" + str(tmpTime.second).zfill(2) + "." + str(tmpTime.microsecond).zfill(6))

    def getDateString(self):
        return str(str(self.startTime.year).zfill(4) + "/" + str(self.startTime.month).zfill(2) + "/" + str(self.startTime.day).zfill(2))

    def getSamplingRateString(self):
        return str(self.samplingRate)

    def getWeekNumberString(self):
        return str(self.weekNumber)

    def getMissingPointsString(self):
        return str(self.missingPoints)

    def getIntTemperatureString(self):
        return str(self.intTemperature)

    def getExtTemperatureString(self):
        return str(self.extTemperature)

    def getChannelRangeString(self,datasetName):
        return str(self.channelRangesPerName[datasetName])

    def clearData(self):
        self.initialize()

    def createDayFolder(self, dataPath, startTime):
        tmpPath = copy.deepcopy(dataPath)
        tmpPath = os.path.join(tmpPath, str(startTime.year).zfill(4))
        CreateDir(tmpPath)
        tmpPath = os.path.join(tmpPath, str(startTime.month).zfill(2))
        CreateDir(tmpPath)
        tmpPath = os.path.join(tmpPath, str(startTime.day).zfill(2))
        CreateDir(tmpPath)
        print(tmpPath)

        return tmpPath

    def createCorruptDataFolder(self, dataPath):
        tmpPath = copy.deepcopy(dataPath)
        tmpPath = os.path.join(tmpPath,CorruptDataSubfolder)
        CreateDir(tmpPath)
        return tmpPath

    def dumpDataToFile(self):
        print("Dumping data to file...")
        # print(data.dataArray[1:10])
        tmpPath = self.createDayFolder(self.acqConfig.dataPath, self.startTime)
        tmpFileName = self.acqConfig.stationName + "_" + \
            str(self.startTime.year).zfill(4) + str(self.startTime.month).zfill(2) + str(self.startTime.day).zfill(2) + "_" \
          + str(self.startTime.hour).zfill(2) + str(self.startTime.minute).zfill(2) + str(self.startTime.second).zfill(2) + ".h5"

        tmpFullFilePath = os.path.join(tmpPath, tmpFileName)

        print("Opening file for write: " + tmpFileName)
        try:
            self.h5fObj = h5py.File(tmpFullFilePath, "w")
            print("File is open successfully.")
        except:
            print("File open error: " + tmpFullFilePath)
            return

        #write software version to the root of the HDF5 data file
        self.h5fObj.attrs["GNOMEAcquirer_version"] = SOFTWARE_VERSION
        self.h5fObj.attrs["LocalFileCreationTime"] = str(datetime.datetime.utcnow().strftime(logDateTimeFormat))
        self.h5fObj.attrs["DataModel"] = "MagneticField_Default"
        self.h5fObj.attrs["DefaultDataset"] = defaultDatasetName
        self.h5fObj.attrs["DefaultMainEquation"] = defaultMagFieldEqAttrName
        self.h5fObj.attrs["DefaultMainEquationVersion"] = "1.0"
        self.h5fObj.attrs["DefaultMainEquationVarName"] = "Magnetic field"


        for dsName in self.acqConfig.attributes:
            #special data analysis has to be done to the dataset of sanity to downsample it and apply thresholding
            if dsName == sanityDatasetName:
                if len(self.dataArray[dsName]) <= BatchesToStoredInSingleFile:
                    self.tmpDataset = self.h5fObj.create_dataset(dsName, data=self.dataArray[dsName],
                                                                 compression="gzip", compression_opts=9)
                else:
                    #The following downsamples the data to be minimum number of points 60, and then limits the size of the array to 60
                    downsampledDS = list((self.dataArray[dsName][::int(np.floor(len(self.dataArray[dsName]) / BatchesToStoredInSingleFile))])[:BatchesToStoredInSingleFile])
                    #threshold by the given threshold in the attributes
                    threshValue = ApplyAttributeType(self.acqConfig.attributes[dsName][SanityChannel_Attr_Thresh_name][Attribute_ValueKey],
                                                     self.acqConfig.attributes[dsName][SanityChannel_Attr_Thresh_name][Attribute_TypeKey])
                    downsampledDS = list(map(lambda x: True if x >= threshValue else False, downsampledDS))
                    #invert if required by user
                    InvCutoff = ApplyAttributeType(
                        self.acqConfig.attributes[dsName][SanityChannel_Attr_InvCutoff_name][Attribute_ValueKey],
                        self.acqConfig.attributes[dsName][SanityChannel_Attr_InvCutoff_name][Attribute_TypeKey])

                    if InvCutoff:
                        downsampledDS = list(not x for x in downsampledDS)

                    #cast to numpy array
                    downsampledDS = np.asarray(downsampledDS,dtype=bool)
                    #save to file
                    self.tmpDataset = self.h5fObj.create_dataset(dsName, data=downsampledDS,
                                                                 compression="gzip", compression_opts=9)

                # SanityChannel_Attr_Thresh_name
                # SanityChannel_Attr_Cutoff_name

            else:
                self.tmpDataset = self.h5fObj.create_dataset(dsName, data=self.dataArray[dsName],
                                                             compression="gzip", compression_opts=9)
            # print("Dataset name: " + str(self.tmpDataset.name))
            for attName in self.acqConfig.attributes[dsName]:
                # print(attName)
                # print(self.datasetsTabs.widget(i).attributesList[attName].Type)
                # print(self.datasetsTabs.widget(i).attributesList[attName].Value)
                try:
                    self.tmpDataset.attrs[attName] = \
                        ApplyAttributeType(
                                           ApplyAttributeReplacement(self.acqConfig.attributes[dsName][attName][Attribute_ValueKey],
                                                      self, self.acqConfig, dsName),
                                           self.acqConfig.attributes[dsName][attName][Attribute_TypeKey])
                except:
                    sys.stderr.write(
                        "Error while trying to cast the parameter \"" + attName + "\" from the attribute list to type " +
                        self.acqConfig.attributes[dsName][attName][Attribute_TypeKey] + ". "
                                                                                   "Make sure that the type you chose for this variable is compatible with the its contents. The contents "
                                                                                   "are: " + ApplyAttributeReplacement(
                            str(self.acqConfig.attributes[dsName][attName][Attribute_ValueKey]), self, self.acqConfig,
                            dsName) + "\n")
                    sys.stderr.flush()

        # write built-in sensor magnetic fields
        self.tmpDataset = self.h5fObj.create_dataset(OutputMagFieldSensorDataset, data=self.auxMagFieldSensorArray,
                                                     compression="gzip", compression_opts=9)
        self.tmpDataset.attrs["Units"] = self.magFieldSenorUnit
        self.tmpDataset.attrs["Range"] = self.magFieldSenorRange

        # write built-in temperature sensor values
        self.h5fObj.create_dataset(InternalTemperaturesDataset, data=self.intTemperatureArray,
                                   compression="gzip", compression_opts=9)
        self.h5fObj.create_dataset(ExternalTemperaturesDataset, data=self.extTemperatureArray,
                                   compression="gzip", compression_opts=9)

        # write errors list to the file
        errorArray = []
        for errorElement in self.errorMsgs:
            QCoreApplication.processEvents()
            errorArray.append(
                [np.str(errorElement[ERROR_IDENTIFIER_SEVERITY]), np.str(errorElement[ERROR_IDENTIFIER_MESSAGE])])
        self.tmpErrorDataset = self.h5fObj.create_dataset("Errors", data=[0],
                                                          compression="gzip", compression_opts=9)
        for i in range(len(errorArray)):
            QCoreApplication.processEvents()
            self.tmpErrorDataset.attrs[OutputPrefix_Error + str(i)] = errorArray[i][0] + ": " + errorArray[i][1]
        self.h5fObj.close()
        print("File written with no errors.")

    #dumps data to file in emergencies (in case data is corrupted)
    def dumpRemainingDataToFile(self,strQueue):
        print("Dumping remaining data to file after discovering a corruption in the data...")
        # print(data.dataArray[1:10])
        hdf5Suffix = ".h5"
        textSuffix = ".txt"
        tmpPath = self.createCorruptDataFolder(self.acqConfig.dataPath)

        #in case there's no data collected yet, don't use the date for the file name as it's unknown
        if self.numOfBatches == 0:
            tmpCurrTime = datetime.datetime.utcnow()
            tmpFileName = self.acqConfig.stationName + "_" + "UnknownDate_WrittenInUTC_" \
                + str(tmpCurrTime.year).zfill(4) + str(tmpCurrTime.month).zfill(2) + str(tmpCurrTime.day).zfill(2) + "_" \
                + str(tmpCurrTime.hour).zfill(2) + str(tmpCurrTime.minute).zfill(2) + str(tmpCurrTime.second).zfill(2)

        else:
            tmpFileName = self.acqConfig.stationName + "_" + \
                str(self.startTime.year).zfill(4) + str(self.startTime.month).zfill(2) + str(self.startTime.day).zfill(2) + "_" \
              + str(self.startTime.hour).zfill(2) + str(self.startTime.minute).zfill(2) + str(self.startTime.second).zfill(2)

        #write the queued strings that are not parsed to hdf5
        with open(os.path.join(tmpPath,tmpFileName)+textSuffix, 'w') as txtFileHandler:
            txtFileHandler.write('\n'.join(strQueue))

        #if there are no batches stored, exit and don't write the hdf5 file
        if self.numOfBatches <= 0:
            print("Batch found to be empty. Skipping...")
            return

        tmpFullFilePath = os.path.join(tmpPath, tmpFileName) + hdf5Suffix

        #fix the end time of the current set of batches


        print("Opening file for write: " + tmpFileName)
        try:
            self.h5fObj = h5py.File(tmpFullFilePath, "w")
            print("File opened for writing...")
        except Exception as e:
            print("File open error: " + tmpFullFilePath + ". Exception says: " + str(e))
        for dsName in self.acqConfig.attributes:
            #special data analysis has to be done to the dataset of sanity to downsample it and apply thresholding
            if dsName == sanityDatasetName:
                if len(self.dataArray[dsName]) <= BatchesToStoredInSingleFile:
                    self.tmpDataset = self.h5fObj.create_dataset(dsName, data=self.dataArray[dsName],
                                                                 compression="gzip", compression_opts=9)
                else:

                    #The following downsamples the data to be minimum number of points 60, and then limits the size of the array to 60
                    downsampledDS = list((self.dataArray[dsName][::int(np.floor(len(self.dataArray[dsName]) / BatchesToStoredInSingleFile))])[:BatchesToStoredInSingleFile])
                    #threshold by the given threshold in the attributes
                    threshValue = ApplyAttributeType(self.acqConfig.attributes[dsName][SanityChannel_Attr_Thresh_name][Attribute_ValueKey],
                                                     self.acqConfig.attributes[dsName][SanityChannel_Attr_Thresh_name][Attribute_TypeKey])
                    downsampledDS = list(map(lambda x: True if x >= threshValue else False, downsampledDS))
                    #invert if required by user
                    InvCutoff = ApplyAttributeType(
                        self.acqConfig.attributes[dsName][SanityChannel_Attr_InvCutoff_name][Attribute_ValueKey],
                        self.acqConfig.attributes[dsName][SanityChannel_Attr_InvCutoff_name][Attribute_TypeKey])

                    if InvCutoff:
                        downsampledDS = list(not x for x in downsampledDS)

                    #cast to numpy array
                    downsampledDS = np.asarray(downsampledDS,dtype=bool)
                    #save to file
                    self.tmpDataset = self.h5fObj.create_dataset(dsName, data=downsampledDS,
                                                                 compression="gzip", compression_opts=9)

                # SanityChannel_Attr_Thresh_name
                # SanityChannel_Attr_Cutoff_name

            else:
                self.tmpDataset = self.h5fObj.create_dataset(dsName, data=self.dataArray[dsName],
                                                             compression="gzip", compression_opts=9)
            print("Dataset name: " + str(self.tmpDataset.name))
            for attName in self.acqConfig.attributes[dsName]:
                # print(attName)
                # print(self.datasetsTabs.widget(i).attributesList[attName].Type)
                # print(self.datasetsTabs.widget(i).attributesList[attName].Value)
                try:
                    self.tmpDataset.attrs[attName] = \
                        ApplyAttributeType(
                                           ApplyAttributeReplacement(self.acqConfig.attributes[dsName][attName][Attribute_ValueKey],
                                                      self, self.acqConfig, dsName),
                                           self.acqConfig.attributes[dsName][attName][Attribute_TypeKey])
                except Exception as e:
                    sys.stderr.write(
                        "Error while trying to cast the parameter \"" + attName + "\" from the attribute list to type " +
                        self.acqConfig.attributes[dsName][attName][Attribute_TypeKey] + ". "
                                                                                   "Make sure that the type you chose for this variable is compatible with the its contents. The contents "
                                                                                   "are: " + ApplyAttributeReplacement(
                            str(self.acqConfig.attributes[dsName][attName][Attribute_ValueKey]), self, self.acqConfig,
                            dsName) + "\n")
                    sys.stderr.write("Exception says: " + str(e) + "\n")
                    sys.stderr.flush()

        # write built-in sensor magnetic fields
        self.tmpDataset = self.h5fObj.create_dataset(OutputMagFieldSensorDataset, data=self.auxMagFieldSensorArray,
                                                     compression="gzip", compression_opts=9)
        self.tmpDataset.attrs["Units"] = self.magFieldSenorUnit
        self.tmpDataset.attrs["Range"] = self.magFieldSenorRange

        # write built-in temperature sensor values
        self.h5fObj.create_dataset(InternalTemperaturesDataset, data=self.intTemperatureArray,
                                   compression="gzip", compression_opts=9)
        self.h5fObj.create_dataset(ExternalTemperaturesDataset, data=self.extTemperatureArray,
                                   compression="gzip", compression_opts=9)

        # write errors list to the file
        errorArray = []
        for errorElement in self.errorMsgs:
            QCoreApplication.processEvents()
            errorArray.append(
                [np.str(errorElement[ERROR_IDENTIFIER_SEVERITY]), np.str(errorElement[ERROR_IDENTIFIER_MESSAGE])])
        self.tmpErrorDataset = self.h5fObj.create_dataset("Errors", data=[0],
                                                          compression="gzip", compression_opts=9)
        for i in range(len(errorArray)):
            QCoreApplication.processEvents()
            self.tmpErrorDataset.attrs[OutputPrefix_Error + str(i)] = errorArray[i][0] + ": " + errorArray[i][1]
        self.h5fObj.close()




#!/usr/bin/python3

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

import serial
from GNOMEAcquirerLibs.AttributeData import *
import re
import os
import sys

if hasattr(sys, 'frozen'):
    ScriptPath = os.path.join(os.path.dirname(sys.executable), "GNOMEAcquirerLibs")
else:
    ScriptPath = os.path.dirname(os.path.realpath(__file__))

ConfigFileName = "Config.bin"

dataset_regex = r"^(?!.*?__)[a-zA-Z\_]+[a-zA-Z0-9\_]*$"
attrib_regex = r"^(?!.*?__)[a-zA-Z\_]+[a-zA-Z0-9\_\(\)]*$"

defaultDatasetName = "MagneticFields"
sanityDatasetName = "SanityChannel"

defaultMagFieldEqAttrName = "MagFieldEq"

AttributeKeyword_StartTime           = "%STARTTIME%"
AttributeKeyword_EndTime             = "%ENDTIME%"
AttributeKeyword_Date                = "%DATE%"
AttributeKeyword_SamplingRate        = "%SAMPLINGRATE%"
AttributeKeyword_SamplingRate_SanityChannel = "%SAMPLINGRATESANITYCHANNEL%"
AttributeKeyword_WeekNumber          = "%WEEKNUMBER%"
# AttributeKeyword_ConversionFactor    = "%CONVERSIONFACTOR%"
# AttributeKeyword_ChannelOffset       = "%OFFSET%"
AttributeKeyword_Altitude            = "%ALTITUDE%"
AttributeKeyword_Latitude            = "%LATITUDE%"
AttributeKeyword_Longitude           = "%LONGITUDE%"
AttributeKeyword_InternalTemperature = "%INTTEMPERATURE%"
AttributeKeyword_ExternalTemperature = "%EXTTEMPERATURE%"
AttributeKeyword_ChannelRange        = "%CHANNELRANGE%"
AttributeKeyword_MissingPoints       = "%MISSINGPOINTS%"

OutputPrefix_Error = "Error_"
OutputMagFieldSensorDataset = "AuxiliaryMagneticFields"
InternalTemperaturesDataset = "InternalTemperatures"
ExternalTemperaturesDataset = "ExternalTemperatures"

ForbiddenDatasetNames = [OutputMagFieldSensorDataset,
                         InternalTemperaturesDataset,
                         ExternalTemperaturesDataset]

Tooltip_DeviceName       = "Devices are selected using their names (COM port names for Windows). Select the device, from which the data acquiring has to be done."
Tooltip_Channel          = "As this dataset will create a data array in the HDF5 file. Which channel should it use from the device?"
Tooltip_ChannelDataType  = "The data is acquired from the device as ASCII text (strings). For the data to become meaningful, it should be cast to some numerical format, such as floating-point numbers. Which format should be used? " \
                           "\nBe aware that you should cast data to something compatible; i.e., you cannot cast 1.3 to an integer."
# Tooltip_ConversionFactor = "Data received from the selected channel will be multiplied by this number."
# Tooltip_ChannelOffset    = "Data received from the selected channel will be added to this number before scaling it with the specified conversion factor."
Tooltip_MagFieldEquation = "The equation, or method, with which all channels have to be combined to produce a magnetic field. \n" \
                           "This can be changed from the attribute MagFieldEq. " \
                           "For example, if you have two datasets, say MagneticFields and Amplitude, you can combine them with any " \
                           "set of mathematical operations to produce a magnetic field. \nFor example: 2*MagneticFields+1/Amplitude[pT]. \n" \
                           "Please be careful about mathematical operations priorities and use brackets. \n" \
                           "This field is necessary for the post processing of the data."
Tooltip_SanityThreshold  = "The " + sanityDatasetName + "'s values are converted to binary numbers while processing the data. This is the threshold value, \n" \
                           "where if the voltage value is greater than this (or equal), the channel value is assigned True, and otherwise is assigned False. You can invert these with the invert option. \n" \
                                                        "Change this from within the attributes."
Tooltip_SanityInvert     = "The " + sanityDatasetName + "'s output is by default True if higher than the threshold and False if lower. If this option is True, it flips that value with a NOT gate. \n" \
                                                        "Change this from within the attributes."

Tooltip_AddAttribute     = "Each dataset contains header information (attributes). You could add custom information to your program."
Tooltip_RemoveAttribute  = "Remove an attribute that you added. You cannot remove default attributes."

Attribute_NameKey  = "Name"
Attribute_TypeKey  = "Type"
Attribute_ValueKey = "Value"

SanityChannel_Attr_Thresh_name = "Threshold(V)"
SanityChannel_Attr_InvCutoff_name = "InvertAfterThreshold"

class DatasetForm(QWidget):
    def __init__(self, datasetName, parent=None):
        super(DatasetForm, self).__init__(parent)

        #locked datasets cannot be deleted
        isLocked = False

        self.name = datasetName

        windowIcon = QIcon(os.path.join(ScriptPath, "icon.png"))
        self.setWindowIcon(windowIcon)

        self.channelLabel = QLabel("Channel:")
        self.channelLine  = QComboBox()
        self.channelLabel.setToolTip(Tooltip_Channel)
        self.channelLine.setToolTip(Tooltip_Channel)

        self.channelDatatypeLabel = QLabel("Channel elements data type: ")
        self.channelDatatypeLine  = QComboBox()
        self.channelDatatypeLine.addItem(GetTypeNameString(AttributeType.Double))
        self.channelDatatypeLine.addItem(GetTypeNameString(AttributeType.Float))
        self.channelDatatypeLine.addItem(GetTypeNameString(AttributeType.Integer64))
        self.channelDatatypeLine.addItem(GetTypeNameString(AttributeType.Integer32))
        self.channelDatatypeLabel.setToolTip(Tooltip_ChannelDataType)
        self.channelDatatypeLine.setToolTip(Tooltip_ChannelDataType)

        # self.channelConversionFactorLabel     = QLabel("Conversion factor: ")
        # self.channelConversionFactorLine      = QLineEdit("1.0")
        # self.channelConversionFactorLine.setValidator(QDoubleValidator(-1e100,1e100,20,self))
        # self.channelConversionFactorUnitLabel = QLabel()
        # self.channelConversionFactorUnitLabel.setAlignment(Qt.AlignCenter)
        # self.channelConversionFactorLabel.setToolTip(Tooltip_ConversionFactor)
        # self.channelConversionFactorLine.setToolTip(Tooltip_ConversionFactor)
        #
        # self.channelOffsetLabel               = QLabel("Channel offset (before conversion): ")
        # self.channelOffsetLine                = QLineEdit("0.0")
        # self.channelOffsetLine.setValidator(QDoubleValidator(-1e100,1e100,20,self))
        # self.channelOffsetUnitLabel               = QLabel("V")
        # self.channelOffsetUnitLabel.setAlignment(Qt.AlignCenter)
        # self.channelOffsetLabel.setToolTip(Tooltip_ChannelOffset)
        # self.channelOffsetLine.setToolTip(Tooltip_ChannelOffset)

        if self.name == defaultDatasetName:
            self.magFieldEquationLabel     = QLabel("Magnetic field equation: ")
            self.magFieldEquationLine      = QLineEdit("")
            self.magFieldEquationLine.setReadOnly(True)
            self.magFieldEquationLabel.setToolTip(Tooltip_MagFieldEquation)
            self.magFieldEquationLine.setToolTip(Tooltip_MagFieldEquation)

        if self.name == sanityDatasetName:
            self.thresholdToBinaryLabel     = QLabel("Threshold to binary: ")
            self.thresholdToBinaryLine      = QLineEdit("")
            self.thresholdToBinaryUnitsLabel = QLabel("Volts")
            self.thresholdToBinaryLine.setReadOnly(True)
            self.thresholdToBinaryLabel.setToolTip(Tooltip_SanityThreshold)
            self.thresholdToBinaryLine.setToolTip(Tooltip_SanityThreshold)
            self.thresholdToBinaryUnitsLabel.setToolTip(Tooltip_SanityThreshold)

            self.thresholdInvertCheckBox = QCheckBox("Invert thresholded value")
            self.thresholdInvertCheckBox.setToolTip(Tooltip_SanityInvert)
            # self.thresholdInvertCheckBox.setDisabled(True)
            self.thresholdInvertCheckBox.clicked.connect(self.preventToggle_InvCutoff)

        self.items = list(map(str,range(1,5)))
        self.channelLine.addItems(self.items)

        #Buttons
        self.addAttribButton    = QPushButton("Add attribute")
        self.removeAttribButton = QPushButton("Remove selected attribute")
        self.addAttribButton.setToolTip(Tooltip_AddAttribute)
        self.removeAttribButton.setToolTip(Tooltip_RemoveAttribute)

        #Buttons connections
        self.addAttribButton.clicked.connect(self.addAttribFunc)
        self.removeAttribButton.clicked.connect(self.deleteAttrib)

        #Groupbox of connection parts
        self.channelSettingsGroupbox  = QGroupBox("Channel settings")
        self.connectionGroupboxLayout = QGridLayout()

        #set the layout of groupbox
        self.channelSettingsGroupbox.setLayout(self.connectionGroupboxLayout)

        #add items to
        self.connectionGroupboxLayout.addWidget(self.channelLabel,1,0,1,1)
        self.connectionGroupboxLayout.addWidget(self.channelLine ,1,1,1,2)
        self.connectionGroupboxLayout.addWidget(self.channelDatatypeLabel,2,0,1,1)
        self.connectionGroupboxLayout.addWidget(self.channelDatatypeLine ,2,1,1,2)
        if self.name == defaultDatasetName:
            self.connectionGroupboxLayout.addWidget(self.magFieldEquationLabel,3,0,1,1)
            self.connectionGroupboxLayout.addWidget(self.magFieldEquationLine ,3,1,1,2)

        if self.name == sanityDatasetName:
            self.connectionGroupboxLayout.addWidget(self.thresholdToBinaryLabel,3,0,1,1)
            self.connectionGroupboxLayout.addWidget(self.thresholdToBinaryLine ,3,1,1,1)
            self.connectionGroupboxLayout.addWidget(self.thresholdToBinaryUnitsLabel ,3,2,1,1)
            self.connectionGroupboxLayout.addWidget(self.thresholdInvertCheckBox,4,0,1,3)

        #columns addresses
        self.AttributesColumns = {}
        self.AttributesColumns[Attribute_TypeKey]  = 0
        self.AttributesColumns[Attribute_NameKey]  = 1
        self.AttributesColumns[Attribute_ValueKey] = 2

        #create the table
        self.attribsTable = QTableWidget()
        self.attribsTable.setColumnCount(len(self.AttributesColumns))
        self.attribNameHeader  = QTableWidgetItem()
        self.attribValueHeader = QTableWidgetItem()
        self.attribTypeHeader  = QTableWidgetItem()
        self.attribNameHeader.setText("Name")
        self.attribValueHeader.setText("Value")
        self.attribTypeHeader.setText("Type")
        self.attribsTable.setHorizontalHeaderItem(self.AttributesColumns[Attribute_NameKey],self.attribNameHeader)
        self.attribsTable.setHorizontalHeaderItem(self.AttributesColumns[Attribute_ValueKey],self.attribValueHeader)
        self.attribsTable.setHorizontalHeaderItem(self.AttributesColumns[Attribute_TypeKey],self.attribTypeHeader)


        #main layout grid
        self.mainLayout = QGridLayout()
        self.mainLayout.addWidget(self.channelSettingsGroupbox, 0, 0, 1, 2)
        self.mainLayout.addWidget(self.removeAttribButton ,1,1,1,1)
        self.mainLayout.addWidget(self.addAttribButton    ,1,0,1,1)
        self.mainLayout.addWidget(self.attribsTable       ,2,0,1,2)

        tableSize = self.attribsTable.size()

        #bind layout to main window
        self.setLayout(self.mainLayout)
        self.setWindowTitle("Data Acquisition")
        self.resize(tableSize.width(),self.height())

        #this next line is for Qt 5
        self.attribsTable.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        #this next line is for Qt 4
        # self.attribsTable.horizontalHeader().setResizeMode(QHeaderView.Stretch)

        # direct connection is important here in order to avoid separating between the effect of changing
        # attributes and responding to them. We want to avoid any delays due to the eventloop
        self.attribsTable.itemChanged.connect(self.itemWasChanged, Qt.DirectConnection)

        # follow up list with all attributes in a table; this helps in restoring value when it's wrong
        self.attributesList = {}

        # when true, items will refuse to changed their values if their values are locked
        self.ImplementLock = True

        # add default attributes
        self.addDefaultAttributes()

        # self.writeConversionFactorUnit()

        self.writeMagFieldEq()

    # def writeConversionFactorUnit(self):
    #     self.channelConversionFactorUnitLabel.setText(self.attributesList["Units"].Value + "/V")

    def preventToggle_InvCutoff(self):
        if SanityChannel_Attr_InvCutoff_name in self.attributesList:

            checkstatus = Qt.Checked if ApplyAttributeType(self.attributesList[SanityChannel_Attr_InvCutoff_name].Value,
                                                                       self.attributesList[SanityChannel_Attr_InvCutoff_name].Type)\
            else Qt.Unchecked
            self.thresholdInvertCheckBox.setChecked(checkstatus)

        QMessageBox.information(self, "This control is only for viewing",
                            "Please change the control from within the attributes.")

    def writeMagFieldEq(self):
        if self.name == defaultDatasetName:
            if defaultMagFieldEqAttrName in self.attributesList:
                self.magFieldEquationLine.setText(self.attributesList[defaultMagFieldEqAttrName].Value)

    def writeSanityChannelThresholdingValues(self):
        if self.name == sanityDatasetName:
            if SanityChannel_Attr_Thresh_name in self.attributesList:

                # print(ApplyAttributeType(self.attributesList[SanityChannel_Attr_Thresh_name].Value,
                #                          self.attributesList[SanityChannel_Attr_Thresh_name].Type))
                self.thresholdToBinaryLine.setText(self.attributesList[SanityChannel_Attr_Thresh_name].Value)
            if SanityChannel_Attr_InvCutoff_name in self.attributesList:
                self.thresholdInvertCheckBox.setChecked(ApplyAttributeType(self.attributesList[SanityChannel_Attr_InvCutoff_name].Value,
                                                                       self.attributesList[SanityChannel_Attr_InvCutoff_name].Type))
    def addDefaultAttributes(self):
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String),"t0"              ,AttributeKeyword_StartTime,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String),"t1"              ,AttributeKeyword_EndTime,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double),"Longitude"       ,AttributeKeyword_Longitude,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double),"Latitude"        ,AttributeKeyword_Latitude,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double),"Altitude"        ,AttributeKeyword_Altitude,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String),"Date"            ,AttributeKeyword_Date,True,True))
        # self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double),"ConversionFactor",AttributeKeyword_ConversionFactor,True,True))
        # self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double),"Offset"          ,AttributeKeyword_ChannelOffset,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String),"ChannelRange"    ,AttributeKeyword_ChannelRange,True,True))
        self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Integer32),"MissingPoints",AttributeKeyword_MissingPoints,True,True))
        if self.name == defaultDatasetName:
            self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String), "Units", "V", True, False))
            self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String),defaultMagFieldEqAttrName,str(defaultDatasetName+"[pT]"),True,False))

        if self.name == sanityDatasetName:
            self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.String), "Units", "", True, True))
            self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Double), SanityChannel_Attr_Thresh_name,"2.5", True, False))
            self.addAttribFunc_fixed(Attribute(GetTypeNameString(AttributeType.Boolean),SanityChannel_Attr_InvCutoff_name,"False", True, False))
            self.addAttribFunc_fixed(
                Attribute(GetTypeNameString(AttributeType.Double), "SamplingRate(Hz)", AttributeKeyword_SamplingRate_SanityChannel,
                          True, True))
        else:
            self.addAttribFunc_fixed(
                Attribute(GetTypeNameString(AttributeType.Double), "SamplingRate(Hz)", AttributeKeyword_SamplingRate,
                          True, True))

    def resetTableItemFromStored(self, item):
        if item.column() == self.AttributesColumns[Attribute_NameKey]:
            self.ImplementLock = False
            self.attribsTable.item(item.row(), item.column()).setText(self.itemname)
            self.ImplementLock = True
        elif item.column() == self.AttributesColumns[Attribute_ValueKey]:
            self.ImplementLock = False
            self.attribsTable.item(item.row(), item.column()).setText(self.attributesList[self.itemname].Value)
            self.ImplementLock = True
        elif item.column() == self.AttributesColumns[Attribute_TypeKey]:
            self.ImplementLock = False
            self.attribsTable.item(item.row(), item.column()).setText(self.attributesList[self.itemname].Type)
            self.ImplementLock = True
        else:
            QMessageBox.information(self, "Severe Error",
                                    "Error: Unknown column type caught when trying to modify item. "
                                    "This shows a serious error.")
            raise AssertionError(
                "Error: Unknown column type caught when trying to modify item. This shows a serious error.")

    def disableControls(self, value=True):
        self.channelSettingsGroupbox.setDisabled(value)
        self.addAttribButton.setDisabled(value)

    def enableControls(self, value=True):
        self.disableControls(not value)

    def itemWasChanged(self, item):
        if self.ImplementLock:
            self.itemname = self.attribsTable.item(item.row(), self.AttributesColumns[Attribute_NameKey]).text()
            #if the value is locked, then restore the value from the attributesList
            if self.attributesList[self.itemname].isValueLocked:
                self.resetTableItemFromStored(item)
                QMessageBox.information(self,"Error","You cannot modify this item. It's locked!")

            #if the value is not locked, change the value
            else:
                if item.column() == self.AttributesColumns[Attribute_NameKey]:
                    self.attributesList[self.attribsTable.item(item.row(), item.column()).text()] = \
                        self.attributesList[self.itemname]
                    del self.attributesList[self.itemname]
                elif item.column() == self.AttributesColumns[Attribute_ValueKey]:
                    """
                    This part updates the values of the attributes that will be saved in the file
                    Since the DatasetForm objects are used as childs of the AcquisitionWindow object,
                    they are accessed through parents. The parents are set when DatasetForm is instantiated.
                    Keep in mind that this must be done only when acquiring data, because changing attributes this way
                    is not compatible with the default mode, including when the program is started.
                    """
                    AcqWindowWidgetReference = self.parent().parent().parent().parent().parent().parent()
                    if AcqWindowWidgetReference.acquisitionRunning:
                        success = AcqWindowWidgetReference.testAttributesCasting()
                        if success:
                            # apply the change to the list of attributes
                            self.attributesList[self.itemname].Value = self.attribsTable.item(item.row(),
                                                                                              item.column()).text()
                            AcqWindowWidgetReference.updateAttributesDuringAcquisition()
                            AcqWindowWidgetReference.saveSettingsToFile(ConfigFileName)

                        else:
                            # if bad value is given, fall back to the stored value
                            self.resetTableItemFromStored(item)
                    else:
                        # if the acquisition is not running, just change the value, as it will be tested when
                        # acquisition starts
                        self.attributesList[self.itemname].Value = self.attribsTable.item(item.row(),
                                                                                          item.column()).text()

                elif item.column() == self.AttributesColumns[Attribute_TypeKey]:
                    self.attributesList[self.itemname].Type = self.attribsTable.item(item.row(),item.column()).text()
                else:
                    QMessageBox.information(self,"Assertion Error","Error: Unknown column type caught when trying to modify item. This shows a serious error.")
                    raise AssertionError("Error: Unknown column type caught when trying to modify item. This shows a serious error.")
        # try:
        #     self.writeConversionFactorUnit()
        # except:
        #     pass

        try:
        # if True:
            self.writeMagFieldEq()
            self.writeSanityChannelThresholdingValues()
        except Exception as e:
            print("Exception was thrown while trying to write values "
                  "from attributes list to text fields. Exception says: " + str(e))

    def addAttribFunc(self):
        attrNameReturn = QInputDialog.getText(self,"Attribute name?","Enter attribute name. Please use CamelCase, and no spaces!")
        attrName = attrNameReturn[0]

        #forbid OutputPrefix_Error as a prefix
        if len(attrName) >= len(OutputPrefix_Error):
            if attrName[:len(OutputPrefix_Error)] == OutputPrefix_Error:
                QMessageBox.information(self,"Error","Attribute name cannot start with \"" + OutputPrefix_Error + "\". This word is reserved.")
                return

        if attrNameReturn[1]:
            if(not attrName == ""):
                attrReturn = QInputDialog.getItem(self,"Attribute name?",
                                                  "Enter attribute name", GetTypesList(), 0, False)
                attrType  = attrReturn[0]
                attrGiven = attrReturn[1]
                if attrGiven:
                    self.addAttribFunc_fixed(Attribute(attrType, attrName.replace(" ", "").replace("\t", ""), None, False, False))
            else:
                QMessageBox.information(self,"Error","Attribute name cannot be empty.")

    #name,value,type,islocked,isvaluelocked
    def addAttribFunc_fixed(self, attribute, modifyIfExists=False):
        if re.match(attrib_regex, attribute.Name) is None:
            QMessageBox.information(self, "Error",
                                    "Attribute name \"" + attribute.Name + "\""
                                    "is invalid. It should start with letters, and can contain numbers"
                                    " and cannot have more than one underscore consecutively.")
            return

        if attribute.Name in self.attributesList:
            if modifyIfExists:
                self.modifyAttribute(attribute.Name, attribute, False)
            return
        self.ImplementLock = False
        nameitem = QTableWidgetItem()
        nameitem.setText(attribute.Name)
        nameitem.setFlags(Qt.ItemIsEditable)
        valitem = QTableWidgetItem()

        if attribute.Value is None:
            attribute.Value = GetTypeDefaultValue_strType(attribute.Type)

        valitem.setText(attribute.Value)
        typeitem = QTableWidgetItem()
        typeitem.setText(attribute.Type)
        typeitem.setFlags(Qt.ItemIsEditable)
        self.attribsTable.setRowCount(self.attribsTable.rowCount()+1)
        self.attribsTable.setItem(self.attribsTable.rowCount()-1, self.AttributesColumns[Attribute_NameKey], nameitem)
        self.attribsTable.setItem(self.attribsTable.rowCount()-1, self.AttributesColumns[Attribute_ValueKey], valitem)
        self.attribsTable.setItem(self.attribsTable.rowCount()-1, self.AttributesColumns[Attribute_TypeKey], typeitem)
        self.scrollTableDown()

        self.attributesList[attribute.Name] = attribute
        self.ImplementLock = True

    def attributeExists(self,attributeName):
        return attributeName in self.attributesList

    def attributeLocked(self, attributeName):
        return self.attributesList[attributeName].isLocked

    def attributeValueLocked(self, attributeName):
        return self.attributesList[attributeName].isValueLocked

    def modifyAttribute(self,name,attribute,changeTypeToo=False,forceModify=False):
        doChange = True
        if self.attributesList[attribute.Name].Type == attribute.Type:
            doChange = True
        else:
            if changeTypeToo:
                doChange = True
            else:
                doChange = False

        if doChange:
            #release lock if forced to be changed
            lockWasOpened = False
            if forceModify:
                #if locked, then state that the lock was opened to relock it later
                if self.ImplementLock:
                    lockWasOpened = True
                self.ImplementLock = False
            self.attributesList[attribute.Name].Value = attribute.Value
            self.attributesList[attribute.Name].Type = attribute.Type
            for i in range(self.attribsTable.rowCount()):
                if self.attribsTable.item(i,self.AttributesColumns[Attribute_NameKey]).text() == attribute.Name:
                    tmpValueItem = QTableWidgetItem()
                    tmpTypeItem = QTableWidgetItem()
                    tmpValueItem.setText(attribute.Value)
                    tmpTypeItem.setText(attribute.Type)
                    tmpTypeItem.setFlags(Qt.ItemIsEditable)
                    self.attribsTable.setItem(i,self.AttributesColumns[Attribute_ValueKey],tmpValueItem)
                    self.attribsTable.setItem(i,self.AttributesColumns[Attribute_TypeKey],tmpTypeItem)

            #restore lock
            if lockWasOpened:
                self.ImplementLock = True


    def deleteAttrib(self):
        selectedRows = self.attribsTable.selectionModel().selectedRows()
        AcqWindowWidgetReference = self.parent().parent().parent().parent().parent().parent()
        if len(selectedRows) > 0:
            for rowObject in selectedRows:
                self.itemname = self.attribsTable.item(rowObject.row(),self.AttributesColumns[Attribute_NameKey]).text()
                if not self.attributesList[self.itemname].isLocked:
                    self.attribsTable.removeRow(rowObject.row())
                    del self.attributesList[self.itemname]
                    AcqWindowWidgetReference.updateAttributesDuringAcquisition()
                    AcqWindowWidgetReference.saveSettingsToFile(ConfigFileName)
                else:
                    QMessageBox.information(self,"Error","This attribute is locked. You can't delete it!")
        else:
            QMessageBox.information(self,"Error","No rows selected. Please select the rows you would like to remove (by clicking on their numbers on the left)")

    def scrollTableDown(self):
        self.attribsTable.verticalScrollBar().setValue(self.attribsTable.verticalScrollBar().maximum())

    def getTypesCombobox(self):
        typitem = QComboBox()
        for i in list(AttributeType):
            typitem.addItem((i.name))
        return typitem

    def closeEvent(self, event):
        while self.connectionGroupboxLayout.count():
            item = self.connectionGroupboxLayout.takeAt(0)
            item.widget().deleteLater()

        while self.mainLayout.count():
            item = self.mainLayout.takeAt(0)
            item.widget().deleteLater()

# if __name__ == '__main__':
#     import sys
#     app = QApplication(sys.argv)
#
#     screen = DatasetForm()
#     screen.show()
#
#     sys.exit(app.exec_())



# import GUI.Dataset
# import GUI.AcquisitionWindow
# import GUI.AttributeData
# import GUI.DataAnalysis
# import GUI.AcquisitionConfig
# import GUI.DataAcquire
# import GUI.DataCollected
# import GUI.DataBatch
# import GUI.Meta



#!/usr/bin/env python

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import sys
import matplotlib
matplotlib.use('Qt5Agg')
matplotlib.rcParams['backend.qt5']='PyQt5'
import pylab

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
import matplotlib.pyplot as plt

import datetime
import copy

class DataAnalysisWidget(QWidget):
    MaxSizeToPlot = 10000
    def __init__(self, parent=None):
        super(DataAnalysisWidget, self).__init__(parent)

        self.mainLayout = QGridLayout()
        self.setLayout(self.mainLayout)

        self.initPlots(1)

    def appendData(self, dataToSave, startTime, samplingRate):
        timeArray = []
        timestep = datetime.timedelta(microseconds=float(10 ** 6) / float(samplingRate))
        t0 = copy.deepcopy(startTime)

        # get a random element to get array size
        for self.tmpDSName in dataToSave:
            pass

        # create time array
        for i in range(len(dataToSave[self.tmpDSName])):
            timeArray.append(mdates.date2num(t0))
            t0 += timestep

        if len(dataToSave) == len(self.dataToPlot):
            self.dataToPlot_timeAxis = self.dataToPlot_timeAxis + timeArray
            for self.tmpDataset in dataToSave:
                self.dataToPlot[self.tmpDataset] = self.dataToPlot[self.tmpDataset] + dataToSave[
                    self.tmpDataset].tolist()
            if len(self.dataToPlot[self.tmpDataset]) > self.MaxSizeToPlot:
                self.dataToPlot_timeAxis = self.dataToPlot_timeAxis[-self.MaxSizeToPlot:]
                for el in self.dataToPlot:
                    self.dataToPlot[el] = self.dataToPlot[el][-self.MaxSizeToPlot:]
        else:
            self.dataToPlot = dataToSave
            for self.tmpDataset in dataToSave:
                self.dataToPlot[self.tmpDataset] = dataToSave[self.tmpDataset].tolist()
            self.dataToPlot_timeAxis = timeArray

    def appendData(self, dataToSave, startTime, samplingRate):
        timeArray = []
        timestep = datetime.timedelta(microseconds=float(10 ** 6) / float(samplingRate))
        t0 = copy.deepcopy(startTime)

        # get a random element to get array size
        for self.tmpDSName in dataToSave:
            pass

        # create time array
        for i in range(len(dataToSave[self.tmpDSName])):
            timeArray.append(mdates.date2num(t0))
            t0 += timestep

        if len(dataToSave) == len(self.dataToPlot):
            self.dataToPlot_timeAxis = self.dataToPlot_timeAxis + timeArray
            for self.tmpDataset in dataToSave:
                self.dataToPlot[self.tmpDataset] = self.dataToPlot[self.tmpDataset] + dataToSave[
                    self.tmpDataset].tolist()
            if len(self.dataToPlot[self.tmpDataset]) > self.MaxSizeToPlot:
                self.dataToPlot_timeAxis = self.dataToPlot_timeAxis[-self.MaxSizeToPlot:]
                for el in self.dataToPlot:
                    self.dataToPlot[el] = self.dataToPlot[el][-self.MaxSizeToPlot:]
        else:
            self.dataToPlot = dataToSave
            for self.tmpDataset in dataToSave:
                self.dataToPlot[self.tmpDataset] = dataToSave[self.tmpDataset].tolist()
            self.dataToPlot_timeAxis = timeArray

    def setData(self, dataToSave, startTime, samplingRate):
        timeArray = []
        timestep = datetime.timedelta(microseconds=float(10 ** 6) / float(samplingRate))
        t0 = copy.deepcopy(startTime)

        # get a random element to get array size
        for self.tmpDSName in dataToSave:
            pass

        # create time array
        for i in range(len(dataToSave[self.tmpDSName])):
            timeArray.append(mdates.date2num(t0))
            t0 += timestep

        self.dataToPlot = dataToSave
        for self.tmpDataset in dataToSave:
            #the following line was changed by removing tolist(), which didn't work when the plotting scheme was changed
            # from passing DataBatch class to passing DataToPlot class
            # self.dataToPlot[self.tmpDataset] = dataToSave[self.tmpDataset].tolist()
            self.dataToPlot[self.tmpDataset] = dataToSave[self.tmpDataset]
        self.dataToPlot_timeAxis = timeArray

    def initPlots(self,numOfPlots):

        self.dataToPlot = []
        self.dataToPlot_timeAxis = []

        # generate the plot
        # prepare figure objects of Matplotlib
        self.gs = gridspec.GridSpec(numOfPlots,1)
        self.fig = Figure(figsize=(600,600), dpi=72, facecolor=(1,1,1), edgecolor=(0,0,0))
        self.ax = []
        try:
            for i in range(numOfPlots):
                self.ax.append(self.fig.add_subplot(self.gs[i]))
                self.ax[-1].plot([])
        except: pass

        # generate the canvas to display the plot
        self.canvas = FigureCanvas(self.fig)

        self.mainLayout.removeWidget(self.canvas)

        # add the plot canvas to a window
        self.mainLayout.addWidget(self.canvas,0,0,1,1)


    def doPlotData(self):

        #generate the plot
        #prepare figure objects of Matplotlib

        self.plotNumCounter = 0
        try:
            for array in sorted(self.dataToPlot):
                QCoreApplication.processEvents()
                self.ax[self.plotNumCounter].cla()
                self.ax[self.plotNumCounter].plot(self.dataToPlot_timeAxis,self.dataToPlot[array])
                self.ax[self.plotNumCounter].set_ylabel(array,size=12)
                self.ax[self.plotNumCounter].ticklabel_format(style='sci', axis='y', scilimits=(0, 0), useOffset=True)
                if (self.plotNumCounter == len(self.dataToPlot)-1):
                    self.ax[self.plotNumCounter].tick_params(labelsize=12)
                    max_xticks = 6
                    xloc = plt.MaxNLocator(max_xticks)
                    self.ax[self.plotNumCounter].xaxis.set_major_locator(xloc)
                    xfmt = mdates.DateFormatter('%H:%M:%S\n%Y-%m-%d')
                    self.ax[self.plotNumCounter].xaxis.set_major_formatter(xfmt)
                    # print(self.dataToPlot[array])
                else:
                    self.ax[self.plotNumCounter].set_xticklabels([])
                    self.ax[self.plotNumCounter].tick_params(
                        axis='x',  # changes apply to the x-axis
                        labelbottom='off')  # labels along the bottom edge are off

                self.plotNumCounter += 1
                QCoreApplication.processEvents()
        except:
            sys.stderr.write("Error while trying to doPlot.\n")

        # generate the canvas to display the plot
        self.canvas.draw()
